{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.Sorts where

import Control.Monad
import Control.Monad.Trans
import Data.IORef
import Data.Generics
import Data.List
import Data.Graph

import Syntax.Pretty
import Syntax.Internal
import Syntax.Values
import Syntax.Sorts

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.Errors
import TypeChecker.Generics
import TypeChecker.Values
import TypeChecker.Abstractions

import Utils.Generics
import Utils.Monads
import Utils.Debug
import Utils.Functions
import Utils.Pairs
import Utils.Names

{--------------------------------------------------------------------------
    Utils
 --------------------------------------------------------------------------}
{-
-- | Change the index of a sort.
setSortIndex :: SortValue -> SortIndex -> SortValue
setSortIndex (VSStar _) i   = VSStar i

-- | Get the index of a sort.
getSortIndex :: SortValue -> SortIndex
getSortIndex (VSStar i)	= i
-}
-- | Generate a fresh sort index
freshSortIndex :: TC SortIndex
freshSortIndex =
    do	poly <- polymorphicLevels
	t <- if poly then return Polymorphic
		     else do x <- fresh "g"
			     return $ Global x
	flip SIMetaVar 0 <$> liftIO (newIORef $ Uninitialized t [] [])
{-
-- | Return the variables mentioned by a constraint
constraintVars :: SortConstraint -> [SortIndexRef]
constraintVars (LessEq s1 s2) = vars s1 `union` vars s2
    where
	vars (SIMetaVar r _)	= [r]
	vars (SIMax ss)		= foldr union [] $ map vars ss

-- | Add a constraint.
addConstraint :: SortIndexRef -> SortConstraint -> TC ()
addConstraint r c = liftIO $
    do	Uninitialized t cs deps <- readIORef r
	writeIORef r $ Uninitialized t (union [c] cs) deps
	mapM_ (addDep r) deps
    where
	addDep r r' =
	    modifyIORef r' $ \ (Uninitialized t cs deps) ->
				Uninitialized t cs (union [r] deps)

-- | Remove all constraints talking about a particular reference.
--   Used when instantiating this reference. First argument: the reference
--   to update (should be 'Uninitialized'. Returns the removed constraints.
removeConstraints :: SortIndexRef -> SortIndexRef -> TC [SortConstraint]
removeConstraints r r' = liftIO $
    do	Uninitialized t xs ds <- readIORef r
	let (ys,zs) = partition (talksAbout r') xs
	writeIORef r (Uninitialized t zs ds)
	return ys
    where
	talksAbout r s = r `elem` constraintVars s

removeDeps :: SortIndexRef -> SortIndexRef -> TC ()
removeDeps r r' = liftIO $
    do	Uninitialized t xs ds <- readIORef r
	writeIORef r $ Uninitialized t xs $ filter (/= r') ds

-- | Simplify constraints. After generalization, so no meta vars should remain.
simplifyConstraints :: [[SortConstraint]] -> [[SortConstraint]]
simplifyConstraints xs = xs --zipWith simplify [0..] xs
    where
-- 	simplify k cs = concatMap (simp k) cs
-- 
-- 	simp k (LessEq s1 s) =
-- 	    case getMaxOccurrence (SIVar k 0) s of
-- 		(Nothing,_)	-> [LessEq s]
-- 		(Just n,ss)
-- 		    | n < 0	-> [LessEq (SIMax ss)]
-- 		    | otherwise	-> []

-- | Get the maximum occurrence of s in s'.
getMaxOccurrence :: SortIndex -> SortIndex -> (Maybe Int, [SortIndex])
getMaxOccurrence s s' =
    case s' of
	SIMax ss    -> getMax' s ss
	_	    -> getMax' s [s']
    where
	getMax' (SIVar k _)	= getMaxV k
	getMax' (SIMetaVar r _) = getMaxMV r

	max' n Nothing	= Just n
	max' n (Just m) = Just $ max n m

	getMaxV k (SIVar k' n : ss)
	    | k == k'	    = (max' n -*- id) $ getMaxV k ss
	getMaxV k (s : ss)  = (id -*- (s:)) $ getMaxV k ss
	getMaxV _ []	    = (Nothing, [])

	getMaxMV r (SIMetaVar r' n : ss)
	    | r == r'	    = (max' n -*- id) $ getMaxMV r ss
	getMaxMV r (s : ss) = (id -*- (s:)) $ getMaxMV r ss
	getMaxMV _ []	    = (Nothing, [])

{--------------------------------------------------------------------------
    Instantiation
 --------------------------------------------------------------------------}
-}
-- | Instantiate polymorphic definition value
instantiate :: PolymorphicDefValue -> TC DefValue
instantiate (PolyDefVal h ar fv xs cs t) =
    return $ DefVal h ar fv [] cs t
--     do	is <- replicateM (length xs) freshSortIndex
-- 	(xs',(cs',t')) <- instantiateSorts is (xs,(cs,t))
-- 	zipWithM_ setConstraints is xs'
-- 	return $ DefVal h ar is cs' t'
--     where
-- 	setConstraints (SIMetaVar r _) cs =
-- 	    mapM_ (addConstraint r) cs
{-
-- | Instantiates the sort index variables in a term
instantiateSorts :: Data a => [SortIndex] -> a -> TC a
instantiateSorts is =
    everywhereButM' defaultStop (mkM iSI `extM` iCl `extM` iV)
    where
	iSI (SIVar k n)	= return $ siAdd (is !! k) n
	iSI si		= return $ si

	iV (VMeta m t)	=
	    do	i <- liftIO $ readIORef m
		case i of
		    Open _	    -> return ()
		    Instantiated v  ->
			do  v' <- instantiateSorts is v
			    liftIO $ writeIORef m $ Instantiated v'
		return $ VMeta m t
	iV v = return v

	iCl cl	= clMapM' (instantiateSorts is) cl

{--------------------------------------------------------------------------
    Generalization
 --------------------------------------------------------------------------}
-}
-- | Generalizes a definition value.
--   Should we ignore the instances?
generalize :: DefValue -> TC PolymorphicDefValue
generalize (DefVal h ar fv _ cs t) =
	return $ PolyDefVal h ar fv [] cs t
-- 	do  ((cs',t'),xs) <- generalizeSorts (cs,t)
-- 	    return $ PolyDefVal h ar xs cs' t'
{-
-- | Generalize sort index variables. Collects all uninstantiated
--   meta variables and replaces them by concrete variables.
generalizeSorts :: Data a => a -> TC (a, [[SortConstraint]])
generalizeSorts x =
    do	(rs,env) <- free x
	verifyConstraints rs
	--rs' <- filterM polymorphic rs
	xs <- getConstraints rs rs env
	return $ (replace rs env x, simplifyConstraints xs)
    where
	getConstraints rs rs' env = mapM constraints rs
	    where
		constraints r =
		    do	Uninitialized _ xs _ <- liftIO $ readIORef r
			return $ replace rs' env xs

	polymorphic r =
	    do	Uninitialized t _ _ <- liftIO $ readIORef r
		case t of
		    Polymorphic	-> return True
		    _		-> return False

	replace :: Data a => [SortIndexRef] -> [(SortIndexRef, SortIndex)] -> a -> a
	replace rs env = everywhereBut' defaultStop
			    (mkT gSI `extT` gCl)
	    where
		gSI si@(SIMetaVar r n) =
		    case findIndex (r==) rs of
			Just k	-> SIVar k n
			_	-> maybe si (replace rs env . flip siAdd n)
				    $ lookup r env
		gSI (SIMax ss)	= siMaxs ss
		gSI s		= s

		gCl = clMap' (replace rs env)

{--------------------------------------------------------------------------
    Unification
 --------------------------------------------------------------------------}

unifySort :: SortValue -> SortValue -> TC ()
unifySort (VSStar i1) (VSStar i2)   = unifySortIx i1 i2
unifySort s1 s2			    = unless (s1 == s2) $ failure $ sortsNotEqualError s1 s2

-- | Unify two sort indices.
unifySortIx :: SortIndex -> SortIndex -> TC ()
unifySortIx i1 i2 =
    do	i1' <- prune i1
	i2' <- prune i2
	case (i1', i2') of
	    (SIMetaVar r1 n, SIMetaVar r2 m)
		| r1 == r2 && n == m	-> return ()
		| r1 == r2		-> failure $ sortInferenceUnificationError i1' i2'
	    (SIMetaVar r n, i2')	-> r =: siAdd i2' (-n)
	    (i1', SIMetaVar _ _)	-> unifySortIx i2' i1'
	    (SIMax xs, SIMax ys)  ->
		do  mapM_ (flip lessEqSortIx i2') xs
		    mapM_ (flip lessEqSortIx i1') ys
	    (i1, i2)		    -> failure $ sortInferenceUnificationError i1' i2'
    where
	prune = pruneSortIndex

-- | Assign a value to a variable
(=:) :: SortIndexRef -> SortIndex -> TC ()
r =: s	=
    case mn of
	Just n	| n > 0	 -> failure $ sortInferenceUnificationError (SIMetaVar r 0) s
		| n == 0 -> mapM_ (flip lessEqSortIx (SIMetaVar r 0)) ss
	_		 ->
	    do	Uninitialized _ xs ds <- liftIO $ readIORef r

		-- Propage the constraints on r to s
		mapM_ propagate xs

		-- Recompute all constraints mentioning r
		mapM_ (recompute r) ds

		liftIO $ writeIORef r (Initialized s)
    where
	(mn, ss) = getMaxOccurrence (SIMetaVar r undefined) s

	propagate c@(LessEq s s')	=
	    do	mapM_ (flip removeDeps r) $ constraintVars c
		lessEqSortIx s s'

	-- Recompute constraints in r' talking about r
	recompute r r' =
	    do	Uninitialized t cs ds <- liftIO $ readIORef r'
		let (cs1,cs2) = partition (talksAbout r) cs
		liftIO $ writeIORef r' $ Uninitialized t cs2 ds
		mapM_ constrain cs1
	    where
		talksAbout r c = r `elem` constraintVars c
		constrain (LessEq s' s) = lessEqSortIx s' s

-- | Generate a constraint saying that a sort is less or equal to
--   another sort.
lessEqSort :: SortValue -> SortValue -> TC ()
lessEqSort (VSStar i1) (VSStar i2)  = lessEqSortIx i1 i2
lessEqSort s1 s2		    = unless (s1 <= s2) $ failure $ sortsNotLeqError s1 s2

-- | Generate a constraint saying that a sort index is less or equal to
--   another sort index.
lessEqSortIx :: SortIndex -> SortIndex -> TC ()
lessEqSortIx i1 i2 =
    do	i1' <- prune i1
	i2' <- prune i2
	case (i1', i2') of
	    (SIMetaVar r1 n, SIMetaVar r2 m)
		| r1 == r2 && n > m -> failure $ customError "Sort index inconsistency detected."
		| r1 == r2	    -> return ()
	    (SIMetaVar r n, i2')    -> addConstraint r $ LessEq (SIMetaVar r 0) $ siAdd i2' (-n)
	    (SIMax is, i)	    -> mapM_ (flip lessEqSortIx i) is
	    _			    -> failure $ customError $ "lessEqSort " ++ show i1' ++ " " ++ show i2'
    where
	prune = pruneSortIndex

-- | Unwraps a sort index. If it is a meta variable it is
--   guaranteed to be 'Uninitialized'. Also compactifies the pointer chain.
pruneSortIndex :: SortIndex -> TC SortIndex
pruneSortIndex i@(SIMetaVar r _) =
    do	x  <- liftIO $ readIORef r
	case x of
	    Uninitialized _ _ _ -> return i
	    Initialized i'  ->
		do  i'' <- pruneSortIndex i'
		    liftIO $ writeIORef r (Initialized i'')
		    return i''
    where

pruneSortIndex (SIMax ss) = siMaxs <$> mapM pruneSortIndex ss
pruneSortIndex i = return i

{--------------------------------------------------------------------------
    Constraint verification
 --------------------------------------------------------------------------}

-- For the next line, haddock reports a Parse error!?
type Path = [(SortIndexRef, Int)] -- REMOVE_IF_HADDOCK

-- | Checks if the top most reference is part of a cycle.
--   Returns the cycle and its length. Assumes that there
--   is no other cycle on the path.
isCycle :: Path -> Maybe ([SortIndexRef], Int)
isCycle ((r,n):path) =
    case break ((r==) . fst) path of
	(p, (_,m):_)	-> Just (r : map fst p, n - m)
	_		-> Nothing

-- | Verifies that the constraints on a set of variables are consistent
--   (i.e. does not contain any negative loops).
--   Invariant: the references should be 'Uninitialized'.
--   It also simplifies the constraints if possible (not yet).
verifyConstraints :: [SortIndexRef] -> TC ()
verifyConstraints rs = mapM_ search $ map (\r -> [(r,0)]) rs
    where
	step :: Path -> IO [Path]
	step ((r,n):p) =
	    do	Uninitialized _ xs _ <- readIORef r
		return $ concatMap (step' p) xs
	    where
		step' p (LessEq _ (SIMetaVar r m)) =
		    [(r, n + m) : p]
		step' _ _ = []

	search :: Path -> TC ()
	search p =
	    case isCycle p of
		Just (rs, n)	-- TODO: Give a good message?
		    | n < 0	-> failure $ customError "Sort index inconsistency detected"
		    | n == 0	-> failure $ customError "here should be optimizations"
		    -- TODO: if n == 0 we can collapse the entire cycle
		_		-> mapM_ search =<< liftIO (step p)

{--------------------------------------------------------------------------
    Free variables
 --------------------------------------------------------------------------}

-- | Compute all free sort variables in an expression. The returned
--   references should all be 'Uninitialized'. It also returns the
--   initialized references and what they are initialized to.
free :: Data a => a -> TC ([SortIndexRef], [(SortIndexRef, SortIndex)])
free = everythingBut plus (defaultStop `extQ` stopDef `extQ` stopSI)
		     (return ([],[]) `mkQ` freeSortIx `extQ` fCl `extQ` fDef)
    where
	stopDef (DefVal _ _ _ _ _) = True
	stopSI :: SortIndex -> Bool
	stopSI _ = True
   
	plus x y =
	    do	(rs1, env1) <- x
		(rs2, env2) <- y
		return (union rs1 rs2, unionBy ((==) **.* fst) env1 env2)

	fCl cl =
	    free =<< clApp cl (vDef "_") -- (DefVal undefined undefined [] (undefined:undefined) undefined))
	fDef (DefVal _ _ is _ _) = free is

freeSortIx :: SortIndex -> TC ([SortIndexRef], [(SortIndexRef, SortIndex)])
freeSortIx i@(SIMetaVar r _) =
    do	i' <- pruneSortIndex i
	case i' of
	    SIMetaVar r' _
		| r == r'   -> return ([r], [])
	    SIMax is	    -> foldr plus (initR r i') $ map freeSortIx is
	    _		    -> initR r i'
    where
	initR r i = return ([], [(r,i)])

	plus x y =
	    do	(rs1, env1) <- x
		(rs2, env2) <- y
		return (union rs1 rs2, unionBy ((==) **.* fst) env1 env2)


freeSortIx _ = return ([],[])

{-
class Free a where
    free :: a -> IO [SortIndexRef]

instance Free SortIndex where
    free i =
	do  i' <- pruneSortIndex i
	    case i' of
		SIMetaVar r -> return [r]
		SISucc i    -> free i
		_	    -> return []

instance Free Closure where
    free (Cl _ f) = free (f $ VTag undefined)

instance (Free a, Free b) => Free (a,b) where
    free (a, b) = liftM2 union (free a) (free b)

instance (Free a, Free b, Free c) => Free (a,b,c) where
    free (a, b, c) = free (a,(b,c))

instance Free a => Free [a] where
    free xs = foldr union [] <$> mapM free xs

instance Free Value where
    free v = case v of
	VId _ t		    -> free t
	VDef _ d	    -> free d
	VApp v1 v2	    -> free (v1,v2)
	VPi v cl t	    -> free (v,cl,t)
	VLam cl t	    -> free (cl, t)
	VSort s		    -> free s
	VTop t		    -> free t
	VPair v1 v2 t	    -> free (v1, v2, t)
	VProj v _	    -> free v
	VSigma _ v _ cl t   -> free (v, cl, t)
	VTag v		    -> return []
	v		    -> error (show v)

instance Free DefValue where
    free (DefVal _ _ is _ _) = free is

instance Free SortValue where
    free s = case s of
	VSUniv _    -> return []
	VSProp	    -> return []
	VSISet	    -> return []
	VSStar i    -> free i

instance Free ClauseValue where
    free (ClauseVal ps v) = free (ps, v)

instance Free TypedPattern where
    free p = case p of
	VarP _ _ t	-> free t
	ConP _ _ t ps	-> free (t,ps)
	PairP _ p q t	-> free (p,q,t)
	WildP _ t	-> free t
-}
-}
